# DO NOT EDIT
# Autogenerated by convert_protofile.py

class VersionDummy:
	class Version:
		V0_1 = 1063369270
		V0_2 = 1915781601
		V0_3 = 1601562686
		V0_4 = 1074539808
		V1_0 = 885177795
	
	class Protocol:
		PROTOBUF = 656407617
		JSON = 2120839367

class Query:
	class QueryType:
		START = 1
		CONTINUE = 2
		STOP = 3
		NOREPLY_WAIT = 4
		SERVER_INFO = 5
	
	class AssocPair:
		pass

class Frame:
	class FrameType:
		POS = 1
		OPT = 2

class Backtrace:
	pass

class Response:
	class ResponseType:
		SUCCESS_ATOM = 1
		SUCCESS_SEQUENCE = 2
		SUCCESS_PARTIAL = 3
		WAIT_COMPLETE = 4
		SERVER_INFO = 5
		CLIENT_ERROR = 16
		COMPILE_ERROR = 17
		RUNTIME_ERROR = 18
	
	class ErrorType:
		INTERNAL = 1000000
		RESOURCE_LIMIT = 2000000
		QUERY_LOGIC = 3000000
		NON_EXISTENCE = 3100000
		OP_FAILED = 4100000
		OP_INDETERMINATE = 4200000
		USER = 5000000
		PERMISSION_ERROR = 6000000
	
	class ResponseNote:
		SEQUENCE_FEED = 1
		ATOM_FEED = 2
		ORDER_BY_LIMIT_FEED = 3
		UNIONED_FEED = 4
		INCLUDES_STATES = 5

class Datum:
	class DatumType:
		R_NULL = 1
		R_BOOL = 2
		R_NUM = 3
		R_STR = 4
		R_ARRAY = 5
		R_OBJECT = 6
		R_JSON = 7
	
	class AssocPair:
		pass

class Term:
	class TermType:
		DATUM = 1
		MAKE_ARRAY = 2
		MAKE_OBJ = 3
		VAR = 10
		JAVASCRIPT = 11
		UUID = 169
		HTTP = 153
		ERROR = 12
		IMPLICIT_VAR = 13
		DB = 14
		TABLE = 15
		GET = 16
		GET_ALL = 78
		EQ = 17
		NE = 18
		LT = 19
		LE = 20
		GT = 21
		GE = 22
		NOT = 23
		ADD = 24
		SUB = 25
		MUL = 26
		DIV = 27
		MOD = 28
		FLOOR = 183
		CEIL = 184
		ROUND = 185
		APPEND = 29
		PREPEND = 80
		DIFFERENCE = 95
		SET_INSERT = 88
		SET_INTERSECTION = 89
		SET_UNION = 90
		SET_DIFFERENCE = 91
		SLICE = 30
		SKIP = 70
		LIMIT = 71
		OFFSETS_OF = 87
		CONTAINS = 93
		GET_FIELD = 31
		KEYS = 94
		VALUES = 186
		OBJECT = 143
		HAS_FIELDS = 32
		WITH_FIELDS = 96
		PLUCK = 33
		WITHOUT = 34
		MERGE = 35
		BETWEEN_DEPRECATED = 36
		BETWEEN = 182
		REDUCE = 37
		MAP = 38
		FOLD = 187
		FILTER = 39
		CONCAT_MAP = 40
		ORDER_BY = 41
		DISTINCT = 42
		COUNT = 43
		IS_EMPTY = 86
		UNION = 44
		NTH = 45
		BRACKET = 170
		INNER_JOIN = 48
		OUTER_JOIN = 49
		EQ_JOIN = 50
		ZIP = 72
		RANGE = 173
		INSERT_AT = 82
		DELETE_AT = 83
		CHANGE_AT = 84
		SPLICE_AT = 85
		COERCE_TO = 51
		TYPE_OF = 52
		UPDATE = 53
		DELETE = 54
		REPLACE = 55
		INSERT = 56
		DB_CREATE = 57
		DB_DROP = 58
		DB_LIST = 59
		TABLE_CREATE = 60
		TABLE_DROP = 61
		TABLE_LIST = 62
		CONFIG = 174
		STATUS = 175
		WAIT = 177
		RECONFIGURE = 176
		REBALANCE = 179
		SYNC = 138
		GRANT = 188
		INDEX_CREATE = 75
		INDEX_DROP = 76
		INDEX_LIST = 77
		INDEX_STATUS = 139
		INDEX_WAIT = 140
		INDEX_RENAME = 156
		SET_WRITE_HOOK = 189
		GET_WRITE_HOOK = 190
		FUNCALL = 64
		BRANCH = 65
		OR = 66
		AND = 67
		FOR_EACH = 68
		FUNC = 69
		ASC = 73
		DESC = 74
		INFO = 79
		MATCH = 97
		UPCASE = 141
		DOWNCASE = 142
		SAMPLE = 81
		DEFAULT = 92
		JSON = 98
		ISO8601 = 99
		TO_ISO8601 = 100
		EPOCH_TIME = 101
		TO_EPOCH_TIME = 102
		NOW = 103
		IN_TIMEZONE = 104
		DURING = 105
		DATE = 106
		TIME_OF_DAY = 126
		TIMEZONE = 127
		YEAR = 128
		MONTH = 129
		DAY = 130
		DAY_OF_WEEK = 131
		DAY_OF_YEAR = 132
		HOURS = 133
		MINUTES = 134
		SECONDS = 135
		TIME = 136
		MONDAY = 107
		TUESDAY = 108
		WEDNESDAY = 109
		THURSDAY = 110
		FRIDAY = 111
		SATURDAY = 112
		SUNDAY = 113
		JANUARY = 114
		FEBRUARY = 115
		MARCH = 116
		APRIL = 117
		MAY = 118
		JUNE = 119
		JULY = 120
		AUGUST = 121
		SEPTEMBER = 122
		OCTOBER = 123
		NOVEMBER = 124
		DECEMBER = 125
		LITERAL = 137
		GROUP = 144
		SUM = 145
		AVG = 146
		MIN = 147
		MAX = 148
		SPLIT = 149
		UNGROUP = 150
		RANDOM = 151
		CHANGES = 152
		ARGS = 154
		BINARY = 155
		GEOJSON = 157
		TO_GEOJSON = 158
		POINT = 159
		LINE = 160
		POLYGON = 161
		DISTANCE = 162
		INTERSECTS = 163
		INCLUDES = 164
		CIRCLE = 165
		GET_INTERSECTING = 166
		FILL = 167
		GET_NEAREST = 168
		POLYGON_SUB = 171
		TO_JSON_STRING = 172
		MINVAL = 180
		MAXVAL = 181
		BIT_AND = 191
		BIT_OR = 192
		BIT_XOR = 193
		BIT_NOT = 194
		BIT_SAL = 195
		BIT_SAR = 196
	
	class AssocPair:
		pass



import six

def decode_utf8(string, encoding='utf-8'):
    if hasattr(string, 'decode'):
        return string.decode(encoding)

    return string

def chain_to_bytes(*strings):
    return b''.join([six.b(string) if isinstance(string, six.string_types) else string for string in strings])



# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.

from __future__ import print_function

import collections
import copy
import distutils.version
import getpass
import inspect
import optparse
import os
import re
import sys
import threading

from rethinkdb import ast, errors, net, query, version

default_batch_size = 200


class RetryQuery(object):

    __connectOptions = None
    __local = None

    def __init__(self, connect_options):
        if 'host' not in connect_options:
            raise AssertionError('Hostname is a required connection parameter')

        if 'port' not in connect_options:
            raise AssertionError('Port number is a required connection parameter')

        connect_options['port'] = int(connect_options['port'])

        if connect_options['port'] <= 0:
            raise AssertionError('Port number can not be less than one')

        self.__connectOptions = copy.deepcopy(connect_options)

        self.__local = threading.local()

    def conn(self, test_connection=True):
        if not hasattr(self.__local, 'connCache'):
            self.__local.connCache = {}

        # check if existing connection is still good
        if os.getpid() in self.__local.connCache and test_connection:
            try:
                ast.expr(0).run(self.__local.connCache[os.getpid()])
            except errors.ReqlError:
                del self.__local.connCache[os.getpid()]

        # cache a new connection
        if not os.getpid() in self.__local.connCache:
            self.__local.connCache[os.getpid()] = net.make_connection(net.DefaultConnection, **self.__connectOptions)

        # return the connection
        return self.__local.connCache[os.getpid()]

    def __call__(self, name, query_str, times=5, run_options=None, test_connection=True):
        # Try a query multiple times to guard against bad connections
        if name is None:
            raise AssertionError('Name can not be none')

        name = str(name)

        if not isinstance(query_str, ast.RqlQuery):
            raise AssertionError('Query must be a ReQL query instead of {value}'.format(value=query_str))

        if not isinstance(times, int) or times < 1:
            raise ValueError('Times must be a positive integer instead of {value}'.format(value=times))

        if run_options is None:
            run_options = {}

        if not isinstance(run_options, dict):
            raise ValueError('Run option must be a dict instead of {value}'.format(value=run_options))

        last_error = None
        test_connection = False

        for _ in range(times):
            try:
                conn = self.conn(test_connection=test_connection)  # we are already guarding for this
            except errors.ReqlError as e:
                last_error = RuntimeError("Error connecting for during '%s': %s" % (name, str(e)))
                test_connection = True
            try:
                return query_str.run(conn, **run_options)
            except (errors.ReqlTimeoutError, errors.ReqlDriverError) as e:
                last_error = RuntimeError("Connection error during '%s': %s" % (name, str(e)))
            # other errors immediately bubble up

        if last_error is not None:
            raise last_error


def print_progress(ratio, indent=0, read=None, write=None):
    total_width = 40
    done_width = min(int(ratio * total_width), total_width)
    sys.stdout.write("\r%(indent)s[%(done)s%(undone)s] %(percent)3d%%%(readRate)s%(writeRate)s\x1b[K" % {
        "indent": " " * indent,
        "done": "=" * done_width,
        "undone": " " * (total_width - done_width),
        "percent": int(100 * ratio),
        "readRate": (" r: %d" % read) if read is not None else '',
        "writeRate": (" w: %d" % write) if write is not None else ''
    })
    sys.stdout.flush()


def check_minimum_version(options, minimum_version='1.6', raise_exception=True):
    minimum_version = distutils.version.LooseVersion(minimum_version)
    version_string = options.retryQuery('get server version', query.db(
        'rethinkdb').table('server_status')[0]['process']['version'])

    matches = re.match(r'(rethinkdb|rebirthdb) (?P<version>(\d+)\.(\d+)\.(\d+)).*', version_string)

    if not matches:
        raise RuntimeError("invalid version string format: %s" % version_string)

    if distutils.version.LooseVersion(matches.group('version')) < minimum_version:
        if raise_exception:
            raise RuntimeError("Incompatible version, expected >= %s got: %s" % (minimum_version, version_string))
        return False
    return True


DbTable = collections.namedtuple('DbTable', ['db', 'table'])
_tableNameRegex = re.compile(r'^(?P<db>\w+)(\.(?P<table>\w+))?$')


class CommonOptionsParser(optparse.OptionParser, object):

    __retryQuery = None
    __connectRegex = re.compile(r'^\s*(?P<hostname>[\w.-]+)(:(?P<port>\d+))?\s*$')

    def format_epilog(self, formatter):
        return self.epilog or ''

    def __init__(self, *args, **kwargs):
        # -- Type Checkers

        def check_tls_option(_, opt_str, value):
            value = str(value)

            if os.path.isfile(value):
                return {'ca_certs': os.path.realpath(value)}
            else:
                raise optparse.OptionValueError('Option %s value is not a file: %r' % (opt_str, value))

        def check_db_table_option(_, _opt_str, value):
            res = _tableNameRegex.match(value)

            if not res:
                raise optparse.OptionValueError('Invalid db or db.table name: %s' % value)
            if res.group('db') == 'rethinkdb':
                raise optparse.OptionValueError('The `rethinkdb` database is special and cannot be used here')

            return DbTable(res.group('db'), res.group('table'))

        def check_positive_int(_, opt_str, value):
            try:
                value = int(value)
                if value < 1:
                    raise ValueError
            except ValueError:
                raise optparse.OptionValueError('%s value must be an integer greater than 1: %s' % (opt_str, value))

            return value

        def check_existing_file(_, opt_str, value):
            if not os.path.isfile(value):
                raise optparse.OptionValueError('%s value was not an existing file: %s' % (opt_str, value))

            return os.path.realpath(value)

        def check_new_file_location(_, opt_str, value):
            try:
                real_value = os.path.realpath(value)
            except Exception:
                raise optparse.OptionValueError('Incorrect value for %s: %s' % (opt_str, value))

            if os.path.exists(real_value):
                raise optparse.OptionValueError('%s value already exists: %s' % (opt_str, value))

            return real_value

        def file_contents(_, opt_str, value):
            if not os.path.isfile(value):
                raise optparse.OptionValueError('%s value is not an existing file: %r' % (opt_str, value))

            try:
                with open(value, 'r') as passwordFile:
                    return passwordFile.read().rstrip('\n')
            except IOError:
                raise optparse.OptionValueError('bad value for %s: %s' % (opt_str, value))

        # -- Callbacks

        def combined_connect_action(obj, opt, value, parser, *args, **kwargs):
            """optparse.takeaction() calls the callback (which this is set as)
               with the following args: self, opt, value, parser *args, **kwargs
            """
            res = self.__connectRegex.match(value)
            if not res:
                raise optparse.OptionValueError("Invalid 'host:port' format: %s" % value)
            if res.group('hostname'):
                parser.values.hostname = res.group('hostname')
            if res.group('port'):
                parser.values.driver_port = int(res.group('port'))

        # -- setup custom Options object

        class CommonOptionChecker(optparse.Option, object):
            TYPES = optparse.Option.TYPES + ('tls_cert', 'db_table', 'pos_int', 'file', 'new_file', 'file_contents')

            TYPE_CHECKER = copy.copy(optparse.Option.TYPE_CHECKER)
            TYPE_CHECKER['tls_cert'] = check_tls_option
            TYPE_CHECKER['db_table'] = check_db_table_option
            TYPE_CHECKER['pos_int'] = check_positive_int
            TYPE_CHECKER['file'] = check_existing_file
            TYPE_CHECKER['new_file'] = check_new_file_location
            TYPE_CHECKER['file_contents'] = file_contents

            ACTIONS = optparse.Option.ACTIONS + ('add_key', 'get_password')
            STORE_ACTIONS = optparse.Option.STORE_ACTIONS + ('add_key', 'get_password')
            TYPED_ACTIONS = optparse.Option.TYPED_ACTIONS + ('add_key',)
            ALWAYS_TYPED_ACTIONS = optparse.Option.ALWAYS_TYPED_ACTIONS + ('add_key',)

            def take_action(self, action, dest, opt, value, values, parser):
                if dest is None:
                    raise AssertionError('Destination can not be none')

                if action == 'add_key':
                    if self.metavar is None:
                        raise AssertionError('Metavar can not be none')

                    values.ensure_value(dest, {})[self.metavar.lower()] = value
                elif action == 'get_password':
                    values[dest] = getpass.getpass('Password for `admin`: ')
                else:
                    super(CommonOptionChecker, self).take_action(action, dest, opt, value, values, parser)

        kwargs['option_class'] = CommonOptionChecker

        # - default description to the module's __doc__
        if 'description' not in kwargs:
            # get calling module
            caller = inspect.getmodule(inspect.stack()[1][0])
            if caller.__doc__:
                kwargs['description'] = caller.__doc__

        # -- add version

        if 'version' not in kwargs:
            kwargs['version'] = "%%prog %s" % version.VERSION

        # -- call super

        super(CommonOptionsParser, self).__init__(*args, **kwargs)

        # -- add common options

        self.add_option(
            "-q",
            "--quiet",
            dest="quiet",
            default=False,
            action="store_true",
            help='suppress non-error messages')
        self.add_option("--debug", dest="debug", default=False, action="store_true", help=optparse.SUPPRESS_HELP)

        connection_group = optparse.OptionGroup(self, 'Connection options')
        connection_group.add_option(
            '-c',
            '--connect',
            dest='driver_port',
            metavar='HOST:PORT',
            help='host and client port of a rethinkdb node to connect (default: localhost:%d)' % net.DEFAULT_PORT,
            action='callback',
            callback=combined_connect_action,
            type='str'
        )
        connection_group.add_option(
            '--driver-port',
            dest='driver_port',
            metavar='PORT',
            help='driver port of a rethinkdb server',
            type='int',
            default=os.environ.get(
                'RETHINKDB_DRIVER_PORT',
                net.DEFAULT_PORT))
        connection_group.add_option(
            '--host-name',
            dest='hostname',
            metavar='HOST',
            help='host and driver port of a rethinkdb server',
            default=os.environ.get(
                'RETHINKDB_HOSTNAME',
                'localhost'))
        connection_group.add_option(
            '-u',
            '--user',
            dest='user',
            metavar='USERNAME',
            help='user name to connect as',
            default=os.environ.get(
                'RETHINKDB_USER',
                'admin'))
        connection_group.add_option(
            '-p',
            '--password',
            dest='password',
            help='interactively prompt for a password for the connection',
            action='get_password')
        connection_group.add_option(
            '--password-file',
            dest='password',
            metavar='PSWD_FILE',
            help='read the connection password from a file',
            type='file_contents')
        connection_group.add_option(
            '--tls-cert',
            dest='ssl',
            metavar='TLS_CERT',
            help='certificate file to use for TLS encryption.',
            type='tls_cert')
        self.add_option_group(connection_group)

    def parse_args(self, *args, **kwargs):
        # - validate options

        connect = True
        if 'connect' in kwargs:
            connect = kwargs['connect']
            del kwargs['connect']

        # - validate ENV variables

        if 'RETHINKDB_DRIVER_PORT' in os.environ:
            driver_port = os.environ['RETHINKDB_DRIVER_PORT']

            if not isinstance(driver_port, int) or driver_port < 1:
                self.error('ENV variable RETHINKDB_DRIVER_PORT is not a useable '
                           'integer: %s'
                           % os.environ['RETHINKDB_DRIVER_PORT'])

        # - parse options

        options, args = super(CommonOptionsParser, self).parse_args(*args, **kwargs)

        # - setup a RetryQuery instance

        self.__retryQuery = RetryQuery(connect_options={
            'host': options.hostname,
            'port': options.driver_port,
            'user': options.user,
            'password': options.password,
            'ssl': options.ssl
        })
        options.retryQuery = self.__retryQuery

        # - test connection

        if connect:
            try:
                options.retryQuery.conn()
            except errors.ReqlError as e:
                self.error('Unable to connect to server: %s' % str(e))

        # -

        return options, args


_interrupt_seen = False


def abort(pools, exit_event):
    global _interrupt_seen

    if _interrupt_seen:
        # second time
        print("\nSecond terminate signal seen, aborting ungracefully")
        for pool in pools:
            for worker in pool:
                worker.terminate()
                worker.join(.1)
    else:
        print("\nTerminate signal seen, aborting")
        _interrupt_seen = True
        exit_event.set()



# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.


__all__ = [
    "ReqlAuthError",
    "ReqlAvailabilityError",
    "ReqlCompileError",
    "ReqlCursorEmpty",
    "ReqlDriverCompileError",
    "ReqlDriverError",
    "ReqlError",
    "ReqlInternalError",
    "ReqlNonExistenceError",
    "ReqlOpFailedError",
    "ReqlOpIndeterminateError",
    "ReqlPermissionError",
    "ReqlQueryLogicError",
    "ReqlResourceLimitError",
    "ReqlRuntimeError",
    "ReqlServerCompileError",
    "ReqlTimeoutError",
    "ReqlUserError",
    "RqlClientError",
    "RqlCompileError",
    "RqlCursorEmpty",
    "RqlDriverError",
    "RqlError",
    "RqlRuntimeError",
    "RqlTimeoutError",
]

import sys

try:
    unicode

    def convertForPrint(inputString):
        if isinstance(inputString, unicode):  # noqa: F821
            encoding = 'utf-8'
            if hasattr(sys.stdout, 'encoding') and sys.stdout.encoding:
                encoding = sys.stdout.encoding
            return inputString.encode(encoding or 'utf-8', 'replace')
        else:
            return str(inputString)
except NameError:
    def convertForPrint(inputString):
        return inputString

try:
    {}.iteritems

    def dict_items(d): return d.iteritems()
except AttributeError:
    def dict_items(d): return d.items()


class ReqlCursorEmpty(Exception):
    def __init__(self):
        super(ReqlCursorEmpty, self).__init__("Cursor is empty.")
        self.message = "Cursor is empty."


RqlCursorEmpty = ReqlCursorEmpty


class ReqlError(Exception):
    def __init__(self, message, term=None, frames=None):
        super(ReqlError, self).__init__(message)
        self.message = message
        self.frames = frames
        if term is not None and frames is not None:
            self.query_printer = QueryPrinter(term, self.frames)

    def __str__(self):
        if self.frames is None:
            return convertForPrint(self.message)
        else:
            return convertForPrint("%s in:\n%s\n%s" % (
                self.message.rstrip("."),
                self.query_printer.print_query(),
                self.query_printer.print_carrots()))

    def __repr__(self):
        return "<%s instance: %s >" % (self.__class__.__name__, str(self))


RqlError = ReqlError


class ReqlCompileError(ReqlError):
    pass


RqlCompileError = ReqlCompileError


class ReqlDriverCompileError(ReqlCompileError):
    pass


class ReqlServerCompileError(ReqlCompileError):
    pass


class ReqlRuntimeError(ReqlError):
    pass


RqlRuntimeError = ReqlRuntimeError


class ReqlQueryLogicError(ReqlRuntimeError):
    pass


class ReqlNonExistenceError(ReqlQueryLogicError):
    pass


class ReqlResourceLimitError(ReqlRuntimeError):
    pass


class ReqlUserError(ReqlRuntimeError):
    pass


class ReqlInternalError(ReqlRuntimeError):
    pass


class ReqlAvailabilityError(ReqlRuntimeError):
    pass


class ReqlOpFailedError(ReqlAvailabilityError):
    pass


class ReqlOpIndeterminateError(ReqlAvailabilityError):
    pass


class ReqlPermissionError(ReqlRuntimeError):
    pass


class ReqlDriverError(ReqlError):
    pass


RqlClientError = ReqlDriverError
RqlDriverError = ReqlDriverError


class ReqlAuthError(ReqlDriverError):
    def __init__(self, msg, host=None, port=None):
        if host is not None and port is not None:
            msg = "Could not connect to {}:{}, {}".format(host, port, msg)
        super(ReqlAuthError, self).__init__(msg)


class _ReqlTimeoutError(ReqlDriverError):
    def __init__(self, host=None, port=None):
        msg = "Operation timed out."
        if host is not None and port is not None:
            msg = "Could not connect to {}:{}, {}".format(host, port, msg)
        super(_ReqlTimeoutError, self).__init__(msg)


try:
    class ReqlTimeoutError(_ReqlTimeoutError, TimeoutError):
        pass
except NameError:
    class ReqlTimeoutError(_ReqlTimeoutError):
        pass

RqlTimeoutError = ReqlTimeoutError


class QueryPrinter(object):
    def __init__(self, root, frames=None):
        self.root = root
        self.frames = list(frames or ())

    def print_query(self):
        return ''.join(self.compose_term(self.root))

    def print_carrots(self):
        return ''.join(self.compose_carrots(self.root, self.frames))

    def compose_term(self, term):
        args = [self.compose_term(a) for a in term._args]
        optargs = {}
        for k, v in dict_items(term.optargs):
            optargs[k] = self.compose_term(v)
        return term.compose(args, optargs)

    def compose_carrots(self, term, frames):
        # This term is the cause of the error
        if len(frames) == 0:
            return ['^' for i in self.compose_term(term)]

        cur_frame = frames[0]
        args = [self.compose_carrots(arg, frames[1:])
                if cur_frame == i else self.compose_term(arg)
                for i, arg in enumerate(term._args)]

        optargs = {}
        for k, v in dict_items(term.optargs):
            if cur_frame == k:
                optargs[k] = self.compose_carrots(v, frames[1:])
            else:
                optargs[k] = self.compose_term(v)

        return [' ' if i != '^' else '^' for i in term.compose(args, optargs)]


# This 'enhanced' tuple recursively iterates over it's elements allowing us to
# construct nested heirarchies that insert subsequences into tree. It's used
# to construct the query representation used by the pretty printer.
class T(object):
    # N.B Python 2.x doesn't allow keyword default arguments after *seq
    #     In Python 3.x we can rewrite this as `__init__(self, *seq, intsp=''`
    def __init__(self, *seq, **opts):
        self.seq = seq
        self.intsp = opts.pop('intsp', '')

    def __iter__(self):
        itr = iter(self.seq)

        try:
            for sub in next(itr):
                yield sub
        except StopIteration:
            return

        for token in itr:
            for sub in self.intsp:
                yield sub

            for sub in token:
                yield sub



# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import imp
import pkg_resources

from rethinkdb import errors, version


# The builtins here defends against re-importing something obscuring `object`.
try:
    import __builtin__ as builtins  # Python 2
except ImportError:
    import builtins  # Python 3


__all__ = ['RethinkDB'] + errors.__all__
__version__ = version.VERSION


class RethinkDB(builtins.object):
    def __init__(self):
        super(RethinkDB, self).__init__()

        from rethinkdb import (
            _dump,
            _export,
            _import,
            _index_rebuild,
            _restore,
            ast,
            query,
            net
        )

        self._dump = _dump
        self._export = _export
        self._import = _import
        self._index_rebuild = _index_rebuild
        self._restore = _restore

        net.Connection._r = self

        for module in (net, query, ast, errors):
            for function_name in module.__all__:
                setattr(self, function_name, getattr(module, function_name))

        self.set_loop_type(None)

    def set_loop_type(self, library=None):
        if library is None:
            self.connection_type = net.DefaultConnection
            return

        # find module file
        manager = pkg_resources.ResourceManager()
        libPath = '%(library)s_net/net_%(library)s.py' % {'library': library}
        if not manager.resource_exists(__name__, libPath):
            raise ValueError('Unknown loop type: %r' % library)

        # load the module
        modulePath = manager.resource_filename(__name__, libPath)
        moduleName = 'net_%s' % library
        moduleFile, pathName, desc = imp.find_module(moduleName, [os.path.dirname(modulePath)])
        module = imp.load_module('rethinkdb.' + moduleName, moduleFile, pathName, desc)

        # set the connection type
        self.connection_type = module.Connection

        # cleanup
        manager.cleanup_resources()

    def connect(self, *args, **kwargs):
        return self.make_connection(self.connection_type, *args, **kwargs)


r = RethinkDB()



#!/usr/bin/env python

# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This file incorporates work covered by the following copyright:
# Copyright 2010-2016 RethinkDB, all rights reserved.


'''`rethinkdb-dump` creates an archive of data from a RethinkDB cluster'''

from __future__ import print_function

import datetime
import os
import platform
import shutil
import sys
import tarfile
import tempfile
import time
import traceback

from rethinkdb import _export, utils_common
from rethinkdb.logger import default_logger

usage = "rethinkdb dump [-c HOST:PORT] [-p] [--password-file FILENAME] [--tls-cert FILENAME] [-f FILE] " \
        "[--clients NUM] [-e (DB | DB.TABLE)]..."
help_epilog = '''
EXAMPLES:
rethinkdb dump -c mnemosyne:39500
  Archive all data from a cluster running on host 'mnemosyne' with a client port at 39500.

rethinkdb dump -e test -f rdb_dump.tar.gz
  Archive only the 'test' database from a local cluster into a named file.

rethinkdb dump -c hades -e test.subscribers -p
  Archive a specific table from a cluster running on host 'hades' which requires a password.'''


def parse_options(argv, prog=None):
    parser = utils_common.CommonOptionsParser(usage=usage, epilog=help_epilog, prog=prog)

    parser.add_option(
        "-f",
        "--file",
        dest="out_file",
        metavar="FILE",
        default=None,
        help='file to write archive to (defaults to rethinkdb_dump_DATE_TIME.tar.gz);\nif FILE is -, use standard '
             'output (note that intermediate files will still be written to the --temp-dir directory)')
    parser.add_option(
        "-e",
        "--export",
        dest="db_tables",
        metavar="DB|DB.TABLE",
        default=[],
        type='db_table',
        help='limit dump to the given database or table (may be specified multiple times)',
        action="append")

    parser.add_option("--temp-dir", dest="temp_dir", metavar="directory", default=None,
                      help='the directory to use for intermediary results')
    parser.add_option(
        "--overwrite-file",
        dest="overwrite",
        default=False,
        help="overwrite -f/--file if it exists",
        action="store_true")
    parser.add_option(
        "--clients",
        dest="clients",
        metavar="NUM",
        default=3,
        help='number of tables to export simultaneously (default: 3)',
        type="pos_int")
    parser.add_option(
        "--read-outdated",
        dest="outdated",
        default=False,
        help='use outdated read mode',
        action="store_true")

    options, args = parser.parse_args(argv)

    # Check validity of arguments
    if len(args) != 0:
        raise parser.error("No positional arguments supported. Unrecognized option(s): %s" % args)

    # Add dump name
    if platform.system() == "Windows" or platform.system().lower().startswith('cygwin'):
        options.dump_name = "rethinkdb_dump_%s" % datetime.datetime.today().strftime(
            "%Y-%m-%dT%H-%M-%S")  # no colons in name
    else:
        options.dump_name = "rethinkdb_dump_%s" % datetime.datetime.today().strftime("%Y-%m-%dT%H:%M:%S")

    # Verify valid output file
    if options.out_file == "-":
        options.out_file = sys.stdout
        options.quiet = True
    elif options.out_file is None:
        options.out_file = os.path.realpath("%s.tar.gz" % options.dump_name)
    else:
        options.out_file = os.path.realpath(options.out_file)

    if options.out_file is not sys.stdout:
        if os.path.exists(options.out_file) and not options.overwrite:
            parser.error("Output file already exists: %s" % options.out_file)
        if os.path.exists(options.out_file) and not os.path.isfile(options.out_file):
            parser.error("There is a non-file at the -f/--file location: %s" % options.out_file)

    # Verify valid client count
    if options.clients < 1:
        raise RuntimeError("Error: invalid number of clients (%d), must be greater than zero" % options.clients)

    # Make sure the temporary directory exists and is accessible
    if options.temp_dir is not None:
        if not os.path.exists(options.temp_dir):
            try:
                os.makedirs(options.temp_dir)
            except OSError:
                parser.error("Could not create temporary directory: %s" % options.temp_dir)
        if not os.path.isdir(options.temp_dir):
            parser.error("Temporary directory doesn't exist or is not a directory: %s" % options.temp_dir)
        if not os.access(options.temp_dir, os.W_OK):
            parser.error("Temporary directory inaccessible: %s" % options.temp_dir)

    return options


def main(argv=None, prog=None):
    options = parse_options(argv or sys.argv[1:], prog=prog)
    try:
        if not options.quiet:
            # Print a warning about the capabilities of dump, so no one is confused (hopefully)
            print("""\
            NOTE: 'rethinkdb-dump' saves data, secondary indexes, and write hooks, but does *not* save
            cluster metadata.  You will need to recreate your cluster setup yourself after
            you run 'rethinkdb-restore'.""")

        try:
            start_time = time.time()
            archive = None

            # -- _export options - need to be kep in-sync with _export

            options.directory = os.path.realpath(tempfile.mkdtemp(dir=options.temp_dir))
            options.fields = None
            options.delimiter = None
            options.format = 'json'

            # -- export to a directory

            if not options.quiet:
                print("  Exporting to temporary directory...")

            try:
                _export.run(options)
            except Exception as exc:
                default_logger.exception(exc)

                if options.debug:
                    sys.stderr.write('\n%s\n' % traceback.format_exc())

                raise Exception("Error: export failed, %s" % exc)

            # -- zip directory

            if not options.quiet:
                print("  Zipping export directory...")

            try:
                if hasattr(options.out_file, 'read'):
                    archive = tarfile.open(fileobj=options.out_file, mode="w:gz")
                else:
                    archive = tarfile.open(name=options.out_file, mode="w:gz")
                for curr, _, files in os.walk(os.path.realpath(options.directory)):
                    for data_file in files:
                        full_path = os.path.join(options.directory, curr, data_file)
                        archive_path = os.path.join(options.dump_name, os.path.relpath(full_path, options.directory))
                        archive.add(full_path, arcname=archive_path)
                        os.unlink(full_path)
            finally:
                if archive:
                    archive.close()

            # --

            if not options.quiet:
                print(
                    "Done (%.2f seconds): %s" %
                    (time.time() -
                     start_time,
                     options.out_file.name if hasattr(
                         options.out_file,
                         'name') else options.out_file))
        except KeyboardInterrupt:
            time.sleep(0.2)
            raise RuntimeError("Interrupted")
        finally:
            if os.path.exists(options.directory):
                shutil.rmtree(options.directory)

    except Exception as ex:
        if options.debug:
            traceback.print_exc()
        print(ex, file=sys.stderr)
        return 1
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2018 RethinkDB
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Wrap logging package to not repeat general logging steps.
"""


import logging
import sys


class DriverLogger(object):
    """
    DriverLogger is a wrapper for logging's debug, info, warning and error functions.
    """

    def __init__(self, level=logging.INFO):
        """
        Initialize DriverLogger

        :param level: Minimum logging level
        :type level: int
        """

        super(DriverLogger, self).__init__()

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(level)

        self.write_to_console = False

    @staticmethod
    def _convert_message(message):
        """
        Convert any message to string.

        :param message: Message to log
        :type message: any
        :return: String representation of the message
        :rtype: str
        """

        return str(message)

    def _print_message(self, level, message):
        if self.write_to_console:
            if level <= logging.WARNING:
                sys.stdout.write(message)
            else:
                sys.stderr.write(message)

    def _log(self, level, message, *args, **kwargs):
        self._print_message(level, message)
        self.logger.log(level, message, *args, **kwargs)

    def debug(self, message):
        """
        Log debug messages.

        :param message: Debug message
        :type message: str
        :rtype: None
        """

        self._log(logging.DEBUG, message)

    def info(self, message):
        """
        Log info messages.

        :param message: Info message
        :type message: str
        :rtype: None
        """

        self._log(logging.INFO, message)

    def warning(self, message):
        """
        Log warning messages.

        :param message: Warning message
        :type message: str
        :rtype: None
        """

        self._log(logging.WARNING, message)

    def error(self, message):
        """
        Log error messages.

        :param message: Error message
        :type message: str
        :rtype: None
        """

        self._log(logging.ERROR, message)

    def exception(self, exc, with_raise=False):
        """
        Log an exception with its traceback and the message if possible.

        :param exc: Exception
        :type exc: str
        :rtype: None
        """

        self._log(logging.ERROR, self._convert_message(exc), exc_info=1)

        if with_raise and isinstance(exc, Exception):
            raise exc


default_logger = DriverLogger()
